@@ -1,5 +1,5 @@
 /*
-  archive.c -- information about an archive
+  Archive.cc -- information about an archive
   Copyright (C) 1999-2014 Dieter Baron and Thomas Klausner
 
   This file is part of ckmame, a program to check rom sets for MAME.
@@ -48,14 +48,15 @@
 #include "ArchiveLibarchive.h"
 #endif
 #include "ArchiveZip.h"
-#include "Detector.h"
+#include "CkmameCache.h"
 #include "CkmameDB.h"
+#include "Detector.h"
 #include "Exception.h"
+#include "Progress.h"
+#include "RomDB.h"
 #include "file_util.h"
 #include "globals.h"
-#include "MemDB.h"
-#include "RomDB.h"
-#include "CkmameCache.h"
+#include "util.h"
 
 #define BUFSIZE 8192
 
@@ -63,9 +64,7 @@
 
 bool Archive::read_only_mode = false;
 
-uint64_t ArchiveContents::next_id = 0;
-std::unordered_map<ArchiveContents::TypeAndName, std::weak_ptr<ArchiveContents>> ArchiveContents::archive_by_name;
-std::unordered_map<uint64_t, ArchiveContentsPtr> ArchiveContents::archive_by_id;
+std::unordered_map<ArchiveContents::TypeAndName, ArchiveContentsPtr> ArchiveContents::archive_by_name;
 
 ArchiveContents::ArchiveContents(ArchiveType type_, std::string name_, filetype_t filetype_, where_t where_, int flags_, std::string filename_extension_) :
     id(0),
@@ -85,16 +84,16 @@ Archive::Archive(ArchiveContentsPtr contents_) :
     name(contents->name),
     filetype(contents->filetype),
     where(contents->where),
-    cache_changed(false),
+    cache_changed(NONE),
     modified(false) {
         changes.resize(files.size());
     }
 
 Archive::Archive(ArchiveType type, const std::string &name_, filetype_t ft, where_t where_, int flags_) : Archive(std::make_shared<ArchiveContents>(type, name_, ft, where_, flags_)) { }
 
-ArchivePtr Archive::open(const ArchiveContentsPtr& contents) {
+ArchivePtr Archive::open(const ArchiveContentsPtr& contents, int flags) {
     ArchivePtr archive;
-    
+
     if (contents->open_archive.expired()) {
         switch (contents->archive_type) {
             case ARCHIVE_LIBARCHIVE:
@@ -104,19 +103,19 @@ ArchivePtr Archive::open(const ArchiveContentsPtr& contents) {
                 return nullptr;
 #endif
                 break;
-                
+
             case ARCHIVE_ZIP:
                 archive = std::make_shared<ArchiveZip>(contents);
                 break;
-                
+
             case ARCHIVE_DIR:
                 archive = std::make_shared<ArchiveDir>(contents);
                 break;
-                
+
             case ARCHIVE_IMAGES:
                 archive = std::make_shared<ArchiveImages>(contents);
                 break;
-                
+
             default:
                 return nullptr;
         }
@@ -128,18 +127,9 @@ ArchivePtr Archive::open(const ArchiveContentsPtr& contents) {
         //printf("# already open %s\n", archive->name.c_str());
         archive = contents->open_archive.lock();
     }
-    
-    return archive;
-}
 
-ArchivePtr Archive::by_id(uint64_t id) {
-    auto contents = ArchiveContents::by_id(id);
-    
-    if (!contents) {
-        return nullptr;
-    }
-    
-    return open(contents);
+    contents->flags |= flags & (ARCHIVE_FL_MASK | ARCHIVE_FL_HASHTYPES_MASK);
+    return archive;
 }
 
 
@@ -159,9 +149,9 @@ void Archive::ensure_valid_archive() {
     }
 
     /* opening the archive failed, rename it and create new one */
-    
+
     auto new_name = ::make_unique_name(name, ".broken");
-    
+
     output.message_verbose("rename broken archive '%s' to '%s'", name.c_str(), new_name.c_str());
     if (!rename_or_move(name, new_name)) {
         throw(Exception("can't rename file")); // TODO: rename_or_move should throw
@@ -187,15 +177,17 @@ int Archive::file_compare_hashes(uint64_t index, const Hashes *hashes) {
 
 bool Archive::file_ensure_hashes(uint64_t idx, size_t detector_id, int hashtypes) {
     auto &file = files[idx];
-    
+
     if (file.has_all_hashes(detector_id, hashtypes)) {
 	return true;
     }
-    
+
     if (file.broken) {
         return false;
     }
 
+    auto progress = Progress::Message("computing hashes for " + name + "/" + file.name);
+
     if (detector_id == 0) {
 	Hashes hashes;
 	hashes.add_types(Hashes::TYPE_ALL);
@@ -208,6 +200,7 @@ bool Archive::file_ensure_hashes(uint64_t idx, size_t detector_id, int hashtypes
 	    f->open();
 	} catch (Exception &e) {
 	    output.error("%s: %s: can't open: %s", name.c_str(), file.name.c_str(), e.what());
+            set_cache_changed(FILES);
 	    file.broken = true;
 	    return false;
 	}
@@ -218,11 +211,13 @@ bool Archive::file_ensure_hashes(uint64_t idx, size_t detector_id, int hashtypes
 
 	case READ_ERROR:
 	    output.error("%s: %s: can't compute hashes: %s", name.c_str(), file.name.c_str(), strerror(errno));
+            set_cache_changed(FILES);
 	    file.broken = true;
 	    return false;
 
 	case CRC_ERROR:
 	    output.error("%s: %s: CRC error: %08x != %08x", name.c_str(), file.name.c_str(), hashes.crc, file.hashes.crc);
+            set_cache_changed(FILES);
 	    file.broken = true;
 	    return false;
 	}
@@ -234,7 +229,8 @@ bool Archive::file_ensure_hashes(uint64_t idx, size_t detector_id, int hashtypes
 	    return false;
 	}
     }
-    cache_changed = true;
+    set_cache_changed(HASHES_ONLY);
+    changes[idx].updated_hashes.insert(detector_id);
 
     return true;
 }
@@ -251,7 +247,7 @@ std::optional<size_t> Archive::file_find_offset(size_t index, size_t size, const
         auto source = get_source(index);
 
         source->open();
-        
+
         auto found = false;
         size_t offset = 0;
         while (offset + size <= file.hashes.size) {
@@ -282,7 +278,7 @@ std::optional<size_t> Archive::file_find_offset(size_t index, size_t size, const
 
 std::optional<size_t> Archive::file_index_by_name(const std::string &filename) const {
     auto index = contents->file_index_by_name(filename);
-    
+
     if (index.has_value() && changes[index.value()].status == Change::DELETED) {
         return {};
     }
@@ -304,7 +300,7 @@ std::string Archive::make_unique_name_in_archive(const std::string &filename) {
 	auto unique = filename.substr(0, filename.length() - ext.length()) + n + ext;
 
 	auto exists = false;
-        
+
         for (auto &file : files) {
             if (file.name == unique) {
                 exists = true;
@@ -322,7 +318,7 @@ std::string Archive::make_unique_name_in_archive(const std::string &filename) {
 
 ArchivePtr Archive::open(const std::string &name, filetype_t filetype, where_t where, int flags) {
     std::string archive_name;
-    
+
     if (name[name.length() - 1] == '/') {
         archive_name = name.substr(0, name.length() - 1);
         flags |= ARCHIVE_FL_TOP_LEVEL_ONLY;
@@ -333,21 +329,21 @@ ArchivePtr Archive::open(const std::string &name, filetype_t filetype, where_t w
     auto contents = ArchiveContents::by_name(filetype, archive_name);
 
     if (contents) {
-        return open(contents);
+        return open(contents, flags);
     }
 
     ArchivePtr archive;
-    
+
     try {
         switch (filetype) {
             case TYPE_ROM:
                 if (configuration.roms_zipped) {
                     auto extension = std::filesystem::path(archive_name).extension();
-                    if (strcasecmp(extension.c_str(), ".zip") == 0) {
+                    if (iequals(extension, ".zip")) {
                         archive = std::make_shared<ArchiveZip>(archive_name, filetype, where, flags);
                     }
 #ifdef HAVE_LIBARCHIVE
-                    else if (strcasecmp(extension.c_str(), ".7z") == 0) {
+                    else if (iequals(extension, ".7z")) {
                         archive = std::make_shared<ArchiveLibarchive>(archive_name, filetype, where, flags);
                     }
 #endif
@@ -359,11 +355,11 @@ ArchivePtr Archive::open(const std::string &name, filetype_t filetype, where_t w
                     archive = std::make_shared<ArchiveDir>(archive_name, filetype, where, flags);
                 }
                 break;
-                
+
             case TYPE_DISK:
                 archive = std::make_shared<ArchiveImages>(archive_name, filetype, where, flags);
                 break;
-                
+
             default:
                 return nullptr;
         }
@@ -371,14 +367,14 @@ ArchivePtr Archive::open(const std::string &name, filetype_t filetype, where_t w
     catch (...) {
         return {};
     }
-    
+
     archive->contents->open_archive = archive;
     archive->contents->flags = ((flags  & (ARCHIVE_FL_MASK | ARCHIVE_FL_HASHTYPES_MASK)) | (read_only_mode ? ARCHIVE_FL_RDONLY : 0));
-    
+
     if (!archive->read_infos() && (flags & ARCHIVE_FL_CREATE) == 0) {
         return {};
     }
-    
+
     ArchiveContents::enter_in_maps(archive->contents);
 
     return archive;
@@ -399,17 +395,17 @@ ArchivePtr Archive::open_toplevel(const std::string &name, filetype_t filetype,
 bool Archive::read_infos() {
     std::vector<File> files_cache;
 
-    cache_changed = false;
-    
+    set_cache_changed(NONE);
+
     contents->read_infos_from_cachedb(&files_cache);
 
     if (contents->cache_id > 0) {
         switch (contents->is_cache_up_to_date()) {
             case -1:
                 return false;
-                
+
             case 0:
-                cache_changed = true;
+                set_cache_changed(FILES);
 		break;
 
 	    case 1:
@@ -424,12 +420,12 @@ bool Archive::read_infos() {
     }
 
     if (!read_infos_xxx()) {
-        cache_changed = true;
+        set_cache_changed(FILES);
 	return false;
     }
 
-    merge_files(files_cache);
     changes.resize(files.size());
+    merge_files(files_cache);
 
     return true;
 }
@@ -446,7 +442,7 @@ int ArchiveContents::is_cache_up_to_date() const {
     if (open_archive.expired()) {
         return 0;
     }
-    
+
     open_archive.lock()->get_last_update();
 
     if (mtime == 0 && size == 0) {
@@ -481,6 +477,7 @@ Archive::GetHashesStatus Archive::get_hashes(ZipSource *source, uint64_t length,
 
             hu.update(buf, n);
             length -= n;
+            Progress::update();
         }
 
         hu.end();
@@ -503,38 +500,46 @@ Archive::GetHashesStatus Archive::get_hashes(ZipSource *source, uint64_t length,
 
 
 void Archive::merge_files(const std::vector<File> &files_cache) {
+    set_cache_changed(NONE);
+
     for (uint64_t i = 0; i < files.size(); i++) {
         auto &file = files[i];
-        
+
         file.filename_extension = contents->filename_extension;
         auto it = std::find_if(files_cache.cbegin(), files_cache.cend(), [&file](const File &file_cache){ return file.name == file_cache.name; });
         if (it != files_cache.cend()) {
             if (file.mtime == (*it).mtime && file.compare_size_hashes(*it)) {
+                if ((file.hashes.get_types() & ~(it->hashes.get_types())) == 0) {
+                    changes[i].updated_hashes.clear();
+                }
+                else {
+                    changes[i].updated_hashes.insert(0);
+                }
                 file.hashes.merge((*it).hashes);
                 file.detector_hashes = it->detector_hashes;
             }
             else {
-                cache_changed = true;
+                set_cache_changed(FILES);
             }
         }
         else {
-            cache_changed = true;
+            set_cache_changed(FILES);
         }
-        
+
         if (want_crc() && !file.hashes.has_type(Hashes::TYPE_CRC)) {
             if (!file_ensure_hashes(i, Hashes::TYPE_ALL)) {
                 file.broken = true;
                 if (it == files_cache.cend() || !(*it).broken) {
-                    cache_changed = true;
+                    set_cache_changed(FILES);
                 }
                 continue;
             }
-            cache_changed = true;
+            set_cache_changed(HASHES_ONLY);
         }
     }
-    
+
     if (files.size() != files_cache.size()) {
-        cache_changed = true;
+        set_cache_changed(FILES);
     }
 }
 
@@ -544,7 +549,7 @@ std::optional<size_t> Archive::file_index(const FileData *file) const {
             return index;
         }
     }
-    
+
     return {};
 }
 
@@ -557,7 +562,7 @@ bool ArchiveContents::read_infos_from_cachedb(std::vector<File> *cached_files) {
         cache_id = -1;
         return true;
     }
-    
+
     cache_db = ckmame_cache->get_db_for_archive(name);
     cache_id = cache_db ? cache_db->get_archive_id(name, filetype) : 0;
 
@@ -574,50 +579,30 @@ bool ArchiveContents::read_infos_from_cachedb(std::vector<File> *cached_files) {
 }
 
 void ArchiveContents::enter_in_maps(const ArchiveContentsPtr &contents) {
-    if (!(contents->flags & ARCHIVE_FL_NOCACHE)) {
-        contents->id = ++next_id;
-        archive_by_id[contents->id] = contents;
-        
-        if (IS_EXTERNAL(contents->where)) {
-            memdb->insert_archive(contents.get());
-        }
-    }
-
     archive_by_name[TypeAndName(contents->filetype, contents->name)] = contents;
 }
 
-ArchiveContentsPtr ArchiveContents::by_id(uint64_t id) {
-    auto it = archive_by_id.find(id);
-    
-    if (it == archive_by_id.end()) {
-        return nullptr;
-    }
-    
-    return it->second;
-}
 
 ArchiveContentsPtr ArchiveContents::by_name(filetype_t filetype, const std::string &name) {
     auto it = archive_by_name.find(TypeAndName(filetype, name));
-    
-    if (it == archive_by_name.end() || it->second.expired()) {
+
+    if (it == archive_by_name.end()) {
         return nullptr;
     }
-    
-    return it->second.lock();
+
+    return it->second;
 }
 
 
 void ArchiveContents::clear_cache() {
     archive_by_name.clear();
-    archive_by_id.clear();
-    next_id = 0;
 }
 
 
 std::optional<size_t> ArchiveContents::file_index_by_name(const std::string &filename) const {
     for (size_t i = 0; i < files.size(); i++) {
         auto &file = files[i];
-        
+
         if (filename == file.name) {
             return i;
         }
@@ -628,28 +613,28 @@ std::optional<size_t> ArchiveContents::file_index_by_name(const std::string &fil
 
 bool Archive::compute_detector_hashes(const std::unordered_map<size_t, DetectorPtr> &detectors) {
     auto got_new_hashes = false;
-    
+
     for (size_t index = 0; index < files.size(); index++) {
         auto &file = files[index];
         std::unordered_map<size_t, DetectorPtr> missing_detectors;
-        
+
         for (const auto &pair : detectors) {
             if (file.detector_hashes.find(pair.first) == file.detector_hashes.end()) {
                 missing_detectors[pair.first] = pair.second;
             }
         }
-        
+
         if (missing_detectors.empty()) {
             continue;
         }
-        
+
         if (compute_detector_hashes(index, missing_detectors)) {
             got_new_hashes = true;
         }
     }
-    
+
     if (got_new_hashes) {
-        cache_changed = true;
+        set_cache_changed(HASHES_ONLY);
     }
     return got_new_hashes;
 }
@@ -657,11 +642,11 @@ bool Archive::compute_detector_hashes(const std::unordered_map<size_t, DetectorP
 
 bool Archive::compute_detector_hashes(size_t index, const std::unordered_map<size_t, DetectorPtr> &detectors) {
     auto &file = files[index];
-    
+
     if (file.broken) {
         return false;
     }
-    
+
     auto data = std::vector<uint8_t>(file.get_size(0));
 
     try {
@@ -675,15 +660,11 @@ bool Archive::compute_detector_hashes(size_t index, const std::unordered_map<siz
     catch (std::exception &e) {
         output.error("%s: %s: can't compute hashes: %s", name.c_str(), file.name.c_str(), e.what());
         file.broken = true;
-            
+
         return false;
     }
 
-    auto ok = Detector::compute_hashes(data, &file, detectors);
-    if (ok) {
-	memdb->update_file(contents.get(), index);
-    }
-    return ok;
+    return Detector::compute_hashes(data, &file, db->detectors, &changes[index].updated_hashes);
 }
 
 
@@ -695,14 +676,14 @@ bool ArchiveContents::has_all_detector_hashes(const std::unordered_map<size_t, D
             }
         }
     }
-    
+
     return true;
 }
 
 
 bool Archive::compare_size_hashes(size_t index, size_t detector_id, const FileData *rom) {
     auto &file = files[index];
-    
+
     auto ok = false;
 
     if (rom->compare_size(file)) {
@@ -721,6 +702,13 @@ bool Archive::compare_size_hashes(size_t index, size_t detector_id, const FileDa
             }
         }
     }
-    
+
     return ok;
 }
+
+void Archive::set_cache_changed(Archive::CacheChange new_changed) {
+    if (new_changed == HASHES_ONLY && cache_changed == FILES) {
+        return;
+    }
+    cache_changed = new_changed;
+}
