@@ -40,6 +40,7 @@
 
 #include "Exception.h"
 #include "RomDB.h"
+#include "StatusDB.h"
 #include "util.h"
 
 bool Configuration::read_config_file(std::vector<toml::table> &config_tables, const std::string &file_name, bool optional) {
@@ -99,16 +100,20 @@ TomlSchema::TypePtr Configuration::section_schema = TomlSchema::table({
     { "extra-directories-append", extra_directories_schema},
     { "fixdat-directory",  TomlSchema::string() },
     { "keep-old-duplicate",  TomlSchema::boolean() },
+    { "mia-games", TomlSchema::string() },
     { "missing-list", TomlSchema::string() },
     { "move-from-extra",  TomlSchema::boolean() },
     { "old-db", TomlSchema::string() },
     { "profiles", TomlSchema::array(TomlSchema::string()) },
     { "report-changes",  TomlSchema::boolean() },
     { "report-correct",  TomlSchema::boolean() },
+    { "report-correct-mia",  TomlSchema::boolean() },
     { "report-detailed",  TomlSchema::boolean() },
     { "report-fixable",  TomlSchema::boolean() },
     { "report-missing",  TomlSchema::boolean() },
+    { "report-missing-mia",  TomlSchema::boolean() },
     { "report-no-good-dump",  TomlSchema::boolean() },
+    { "report-status",  TomlSchema::boolean() },
     { "report-summary",  TomlSchema::boolean() },
     { "rom-directory", TomlSchema::string() },
     { "rom-db", TomlSchema::string() },
@@ -116,13 +121,18 @@ TomlSchema::TypePtr Configuration::section_schema = TomlSchema::table({
     { "saved-directory", TomlSchema::string() },
     { "sets", TomlSchema::array(TomlSchema::string()) },
     { "sets-file", TomlSchema::string() },
+    { "status-db", TomlSchema::string() },
+    { "status-db-keep-days", TomlSchema::alternatives({TomlSchema::integer(),TomlSchema::constant("all")}, "integer or 'all'") },
+    { "status-db-keep-runs", TomlSchema::alternatives({TomlSchema::integer(),TomlSchema::constant("all")}, "integer or 'all'") },
     { "unknown-directory", TomlSchema::string() },
     { "update-database",  TomlSchema::boolean() },
     { "use-central-cache-directory", TomlSchema::boolean() },
     { "use-description-as-name",  TomlSchema::boolean() },
     { "use-temp-directory",  TomlSchema::boolean() },
     { "use-torrentzip",  TomlSchema::boolean() },
-    { "verbose",  TomlSchema::boolean() }
+    { "verbose",  TomlSchema::boolean() },
+    { "warn-file-known",  TomlSchema::boolean() },
+    { "warn-file-unknown",  TomlSchema::boolean() }
 }, {});
 
 
@@ -142,24 +152,29 @@ std::vector<Commandline::Option> Configuration::commandline_options = {
     Commandline::Option("fixdat-directory", "directory", "create fixdats in directory"),
     Commandline::Option("keep-old-duplicate", "keep files in ROM set that are also in old ROMs"),
     Commandline::Option("list-sets", "list all known sets"),
+    Commandline::Option("mia-games", "file", "read list of MIA games from file"),
     Commandline::Option("missing-list", "file", "write list of missing games to file"),
     Commandline::Option("move-from-extra", 'j', "remove used files from extra directories"),
     Commandline::Option("no-complete-games-only", "keep partial games in ROM set (default)"),
     Commandline::Option("no-create-fixdat", "don't create fixdat (default)"),
     Commandline::Option("no-report-changes", "don't report changes to correct and missing lists (default)"),
     Commandline::Option("no-report-correct", "don't report status of ROMs that are correct (default)"),
+    Commandline::Option("no-report-correct-mia", "don't report status of ROMs that are correct but marked as mia in ROM database (default)"),
     Commandline::Option("no-report-detailed", "don't report status of every ROM (default)"),
     Commandline::Option("no-report-fixable", "don't report status of ROMs that can be fixed"),
-    Commandline::Option("no-report-missing", "don't report status of ROMs that are missing"),
+    Commandline::Option("no-report-missing", "don't report status of ROMs that are missing and not marked as mia in ROM database"),
+    Commandline::Option("no-report-missing-mia", "don't report status of ROMs that are missing and marked as mia in ROM database (default)"),
     Commandline::Option("no-report-no-good-dump", "don't report status of ROMs for which no good dump exists (default)"),
     Commandline::Option("no-report-summary", "don't print summary of ROM set status (default)"),
     Commandline::Option("no-update-database", "don't update ROM database (default)"),
     Commandline::Option("old-db", 'O', "dbfile", "use database dbfile for old ROMs"),
     Commandline::Option("report-changes", "report changes to correct and missing lists"),
-    Commandline::Option("report-correct", 'c', "report status of ROMs that are correct"),
+    Commandline::Option("report-correct", 'c', "report status of ROMs that are correct but marked as mia in ROM database"),
+    Commandline::Option("report-correct-mia", "report status of ROMs that are correct"),
     Commandline::Option("report-detailed", "report status of every ROM"),
     Commandline::Option("report-fixable", "report status of ROMs that can be fixed (default)"),
-    Commandline::Option("report-missing", "report status of ROMs that are missing (default)"),
+    Commandline::Option("report-missing", "report status of ROMs that are missing and not marked as mia in ROM database (default)"),
+    Commandline::Option("report-missing-mia", "report status of ROMs that are missing and marked as mia in ROM database"),
     Commandline::Option("report-no-good-dump", "don't suppress reporting status of ROMs for which no good dump exists"),
     Commandline::Option("report-summary", "print summary of ROM set status"),
     Commandline::Option("rom-db", 'D', "dbfile", "use ROM database dbfile"),
@@ -167,12 +182,18 @@ std::vector<Commandline::Option> Configuration::commandline_options = {
     Commandline::Option("roms-unzipped", "ROMs are files on disk, not contained in zip archives"),
     Commandline::Option("saved-directory", "directory", "save needed ROMs in directory (default: 'saved')"),
     Commandline::Option("set", "pattern", "check ROM sets matching pattern"),
+    Commandline::Option("no-status-db", "don't save status in dbfile"),
+    Commandline::Option("status-db", "dbfile", "use status in dbfile (default: '.ckmame-status.db')"),
+    Commandline::Option("status-db-keep-days", "n", "remove runs older than n days from status db, 'all' for no time limit (default: 'all')"),
+    Commandline::Option("status-db-keep-runs", "n", "keep only n most recent runs in status db, 'all' for no run limit (default: 2)"),
     Commandline::Option("unknown-directory", "directory", "save unknown files in directory (default: 'unknown')"),
     Commandline::Option("update-database", "update ROM database if dat files changed"),
     Commandline::Option("use-description-as-name", "use description as name of games in ROM database"),
     Commandline::Option("use-temp-directory", 't', "create output in temporary directory, move when done"),
     Commandline::Option("use-torrentzip", "use TORRENTZIP format for zip archives in ROM set"),
-    Commandline::Option("verbose", 'v', "print fixes made")
+    Commandline::Option("verbose", 'v', "print fixes made"),
+    Commandline::Option("warn-file-known", "report status of extra files that are known (default)"),
+    Commandline::Option("warn-file-unknown", "report status of extra files that are unknown (default)")
 };
 
 
@@ -183,12 +204,17 @@ std::unordered_map<std::string, std::string> Configuration::option_to_variable =
     { "no-create-fixdat", "create_fixdat" },
     { "no-report-changes", "report_changes" },
     { "no-report-correct", "report_correct" },
+    { "no-report-correct-mia", "report_correct_mia" },
     { "no-report-detailed", "report_detailed" },
     { "no-report-fixable", "report_fixable" },
     { "no-report-missing", "report_missing" },
+    { "no-report-missing-mia", "report_missing_mia" },
+    { "no-report-status", "report_status" },
     { "no-report-summary", "report_summary" },
     { "no-report-no-good-dump", "report_no_good_dump" },
     { "no-update-database", "update_database" },
+    { "no-warn-file-known", "warn-file-known"},
+    { "no-warn-file-unknown", "warn-file-unknown"},
     { "roms-unzipped", "roms_zipped" }
 };
 
@@ -240,15 +266,19 @@ void Configuration::reset() {
     create_fixdat = false;
     delete_unknown_pattern = "";
     keep_old_duplicate = false;
+    mia_games = "";
     missing_list = "";
     move_from_extra = false;
     old_db = RomDB::default_old_name();
     report_correct = false;
+    report_correct_mia = false;
     report_changes = false;
     report_detailed = false;
     report_fixable = true;
     report_missing = true;
+    report_missing_mia = false;
     report_no_good_dump = false;
+    report_status = true;
     report_summary = false;
     rom_db = RomDB::default_name();
     rom_directory = "roms";
@@ -257,6 +287,8 @@ void Configuration::reset() {
     if (!set.empty()) {
         saved_directory += "/" + set;
     }
+    status_db = StatusDB::default_name();
+    status_db_keep_runs = 2;
     unknown_directory = "unknown";
     update_database = false;
     use_central_cache_directory = false;
@@ -404,6 +436,9 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "keep-old-duplicate") {
             keep_old_duplicate = true;
         }
+        else if (option.name == "mia-games") {
+            mia_games = option.argument;
+        }
         else if (option.name == "missing-list") {
             missing_list = option.argument;
         }
@@ -422,6 +457,9 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "no-report-correct") {
             report_correct = false;
         }
+        else if (option.name == "no-report-correct-mia") {
+            report_correct_mia = false;
+        }
         else if (option.name == "no-report-detailed") {
             report_detailed = false;
         }
@@ -431,15 +469,27 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "no-report-missing") {
             report_missing = false;
         }
+        else if (option.name == "no-report-missing-mia") {
+            report_missing_mia = false;
+        }
         else if (option.name == "no-report-summary") {
             report_summary = false;
         }
         else if (option.name == "no-report-no-good-dump") {
             report_no_good_dump = false;
         }
+        else if (option.name == "no-status-db") {
+            status_db = "none";
+        }
         else if (option.name == "no-update-database") {
             update_database = false;
         }
+        else if (option.name == "no-warn-file-known") {
+            warn_file_known = false;
+        }
+        else if (option.name == "no-warn-file-unknown") {
+            warn_file_unknown = false;
+        }
         else if (option.name == "old-db") {
             old_db = option.argument;
         }
@@ -449,6 +499,9 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "report-correct") {
             report_correct = true;
         }
+        else if (option.name == "report-correct-mia") {
+            report_correct_mia = true;
+        }
         else if (option.name == "report-detailed") {
             report_detailed = true;
         }
@@ -458,12 +511,18 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "report-missing") {
             report_missing = true;
         }
-        else if (option.name == "report-summary") {
-            report_summary = true;
+        else if (option.name == "report-missing-mia") {
+            report_missing_mia = true;
         }
         else if (option.name == "report-no-good-dump") {
             report_no_good_dump = true;
         }
+        else if (option.name == "report-status") {
+            report_status = true;
+        }
+        else if (option.name == "report-summary") {
+          report_summary = true;
+        }
         else if (option.name == "rom-db") {
             rom_db = option.argument;
         }
@@ -476,6 +535,25 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "saved-directory") {
             saved_directory = option.argument;
         }
+        else if (option.name == "status-db") {
+            status_db = option.argument;
+        }
+        else if (option.name == "status-db-keep-days") {
+            if (option.argument == "all") {
+                status_db_keep_days = {};
+            }
+            else {
+                status_db_keep_days = atoi(option.argument.c_str()); // TODO: better conversion with error checking.
+            }
+        }
+        else if (option.name == "status-db-keep-runs") {
+            if (option.argument == "all") {
+                status_db_keep_runs = {};
+            }
+            else {
+                status_db_keep_runs = atoi(option.argument.c_str()); // TODO: better conversion with error checking.
+            }
+        }
         else if (option.name == "unknown-directory") {
             unknown_directory = option.argument;
         }
@@ -494,6 +572,12 @@ void Configuration::prepare(const std::string &current_set, const ParsedCommandl
         else if (option.name == "verbose") {
             verbose = true;
         }
+        else if (option.name == "warn-file-known") {
+            warn_file_known = true;
+        }
+        else if (option.name == "warn-file-unknown") {
+            warn_file_unknown = true;
+        }
     }
 }
 
@@ -539,19 +623,26 @@ void Configuration::merge_config_table(const toml::table *table_pointer) {
     merge_extra_directories(table, "extra-directories-append", true);
     set_string(table, "fixdat-directory", fixdat_directory);
     set_bool(table, "keep-old-duplicate", keep_old_duplicate);
+    set_string(table, "mia-games", mia_games);
     set_string(table, "missing-list", missing_list);
     set_bool(table, "move-from-extra", move_from_extra);
     set_string(table, "old-db", old_db);
     set_bool(table, "report-changes", report_changes);
     set_bool(table, "report-correct", report_correct);
+    set_bool(table, "report-correct-mia", report_correct_mia);
     set_bool(table, "report-detailed", report_detailed);
     set_bool(table, "report-fixable", report_fixable);
     set_bool(table, "report-missing", report_missing);
+    set_bool(table, "report-missing-mia", report_missing_mia);
+    set_bool(table, "report-status", report_status);
     set_bool(table, "report-summary", report_summary);
     set_bool(table, "report-no-good-dump", report_no_good_dump);
     set_string(table, "rom-directory", rom_directory);
     set_bool(table, "roms-zipped", roms_zipped);
     set_string(table, "saved-directory", saved_directory);
+    set_string(table, "status-db", status_db);
+    set_integer_or_all(table, "status-db-keep-days", status_db_keep_days);
+    set_integer_or_all(table, "status-db-keep-runs", status_db_keep_runs);
     set_string(table, "unknown-directory", unknown_directory);
     set_bool(table, "update-database", update_database);
     set_bool(table, "use-central-cache-directory", use_central_cache_directory);
@@ -559,6 +650,8 @@ void Configuration::merge_config_table(const toml::table *table_pointer) {
     set_bool(table, "use-temp-directory", use_temp_directory);
     set_bool(table, "use-torrentzip", use_torrentzip);
     set_bool(table, "verbose", verbose);
+    set_bool(table, "warn-file-known", warn_file_known);
+    set_bool(table, "warn-file-unknown", warn_file_unknown);
 }
 
 
@@ -588,6 +681,32 @@ void Configuration::set_bool_optional(const toml::table &table, const std::strin
     }
 }
 
+void Configuration::set_integer(const toml::table &table, const std::string &name, int &variable) {
+    auto value = table[name].value<int>();
+    if (value) {
+        variable = *value;
+    }
+}
+
+void Configuration::set_integer_optional(const toml::table &table, const std::string &name, std::optional<int> &variable) {
+    auto value = table[name].value<int>();
+    if (value) {
+        variable = *value;
+    }
+}
+
+void Configuration::set_integer_or_all(const toml::table &table, const std::string &name, std::optional<int> &variable) {
+    auto string_value = table[name].value<std::string>();
+    if (string_value && string_value == "all") {
+        variable = {};
+    }
+    else {
+        auto value = table[name].value<int>();
+        if (value) {
+            variable = *value;
+        }
+    }
+}
 
 void Configuration::set_string(const toml::table &table, const std::string &name, std::string &variable) {
     auto value = table[name].value<std::string>();
