@@ -1,5 +1,5 @@
 /*
-  romdb.c -- mame.db sqlite3 data base
+  RomDB.cc -- mame.db sqlite3 data base
   Copyright (C) 2013-2014 Dieter Baron and Thomas Klausner
 
   This file is part of ckmame, a program to check rom sets for MAME.
@@ -41,7 +41,7 @@ std::unique_ptr<RomDB> old_db;
 
 const DB::DBFormat RomDB::format = {
     0x0,
-    3,
+    4,
     "\
 create table dat (\n\
     dat_idx integer primary key,\n\
@@ -54,7 +54,7 @@ create table dat (\n\
 create table game (\n\
     game_id integer primary key autoincrement,\n\
     name text not null,\n\
-        parent text,\n\
+    parent text,\n\
     description text,\n\
     dat_idx integer not null\n\
 );\n\
@@ -72,6 +72,8 @@ create table file (\n\
     crc integer,\n\
     md5 binary,\n\
     sha1 binary,\n\
+    sha256 binary,\n\
+    missing integer not null,\n\
     primary key (game_id, file_type, file_idx)\n\
 );\n\
 create index file_game_type on file (game_id, file_type);\n\
@@ -94,7 +96,13 @@ create table test (\n\
     result integer not null,\n\
     primary key (rule_idx, test_idx)\n\
 );\n",
-    {}
+    {
+        {MigrationVersions(3, 4), "\
+alter table file add column sha256 binary;\n\
+alter table file add column missing integer not null default 0;\n\
+"}
+    }
+
 };
 
 const std::string RomDB::init2_sql = "\
@@ -110,7 +118,7 @@ std::unordered_map<int, std::string> RomDB::queries = {
     {  DELETE_GAME, "delete from game where game_id = :game_id" },
     {  INSERT_DAT_DETECTOR, "insert into dat (dat_idx, name, author, version) values (-1, :name, :author, :version)" },
     {  INSERT_DAT, "insert into dat (dat_idx, name, description, version) values (:dat_idx, :name, :description, :version)" },
-    {  INSERT_FILE, "insert into file (game_id, file_type, file_idx, name, merge, status, location, size, crc, md5, sha1) values (:game_id, :file_type, :file_idx, :name, :merge, :status, :location, :size, :crc, :md5, :sha1)" },
+    {  INSERT_FILE, "insert into file (game_id, file_type, file_idx, name, merge, status, location, size, crc, md5, sha1, sha256, missing) values (:game_id, :file_type, :file_idx, :name, :merge, :status, :location, :size, :crc, :md5, :sha1, :sha256, :missing)" },
     {  INSERT_GAME, "insert into game (name, description, dat_idx, parent) values (:name, :description, :dat_idx, :parent)" },
     {  INSERT_RULE, "insert into rule (rule_idx, start_offset, end_offset, operation) values (:rule_idx, :start_offset, :end_offset, :operation)" },
     {  INSERT_TEST, "insert into test (rule_idx, test_idx, type, offset, size, mask, value, result) values (:rule_idx, :test_idx, :type, :offset, :size, :mask, :value, :result)" },
@@ -118,15 +126,17 @@ std::unordered_map<int, std::string> RomDB::queries = {
     {  QUERY_DAT_DETECTOR, "select name, author, version from dat where dat_idx = -1" },
     {  QUERY_DAT, "select name, description, version from dat where dat_idx >= 0 order by dat_idx" },
     {  QUERY_FILE_FBN, "select g.name, f.file_idx from game g, file f where f.game_id = g.game_id and f.file_type = :file_type and f.name = :name" },
-    {  QUERY_FILE, "select name, merge, status, location, size, crc, md5, sha1 from file where game_id = :game_id and file_type = :file_type order by file_idx" },
+    {  QUERY_FILE, "select name, merge, status, location, size, crc, md5, sha1, sha256, missing from file where game_id = :game_id and file_type = :file_type order by file_idx" },
     {  QUERY_GAME_ID, "select game_id from game where name = :name" },
     {  QUERY_GAME, "select game_id, description, dat_idx, parent from game where name = :name" },
     {  QUERY_HAS_DISKS, "select file_idx from file where file_type = 1 limit 1" },
     {  QUERY_HASH_TYPE_CRC, "select name from file where file_type = :file_type and crc not null limit 1" },
     {  QUERY_HASH_TYPE_MD5, "select name from file where file_type = :file_type and md5 not null limit 1" },
     {  QUERY_HASH_TYPE_SHA1, "select name from file where file_type = :file_type and sha1 not null limit 1" },
+    {  QUERY_HASH_TYPE_SHA256, "select name from file where file_type = :file_type and sha256 not null limit 1" },
     {  QUERY_LIST_DISK, "select distinct name from file where file_type = 1 order by name" },
     {  QUERY_LIST_GAME, "select name from game order by name" },
+    {  QUERY_LIST_MIA, "select game.name as name from game, file where file.game_id = game.game_id and file.missing group by game.name order by game.name" },
     {  QUERY_PARENT_BY_NAME, "select parent from game where name = :name" },
     {  QUERY_PARENT, "select parent from game where game_id = :game_id" },
     {  QUERY_RULE, "select rule_idx, start_offset, end_offset, operation from rule order by rule_idx" },
@@ -138,11 +148,11 @@ std::unordered_map<int, std::string> RomDB::queries = {
 };
 
 std::unordered_map<int, std::string> RomDB::parameterized_queries = {
-   {  QUERY_FILE_FBH, "select g.name as game_name, g.dat_idx, f.file_idx, f.name, f.size, f.crc, f.md5, f.sha1 from game g, file f where f.game_id = g.game_id and f.file_type = :file_type and f.status <> :status @HASH@" },
+   {  QUERY_FILE_FBH, "select g.name as game_name, g.dat_idx, f.file_idx, f.name, f.size, f.crc, f.md5, f.sha1, f.sha256 from game g, file f where f.game_id = g.game_id and f.file_type = :file_type and f.status <> :status @HASH@" },
 
 };
 
-const RomDB::Statement RomDB::query_hash_type[] = { QUERY_HASH_TYPE_CRC, QUERY_HASH_TYPE_MD5, QUERY_HASH_TYPE_SHA1 };
+const RomDB::Statement RomDB::query_hash_type[] = { QUERY_HASH_TYPE_CRC, QUERY_HASH_TYPE_MD5, QUERY_HASH_TYPE_SHA1, QUERY_HASH_TYPE_SHA256 };
 
 
 void RomDB::init2() {
@@ -438,6 +448,7 @@ void RomDB::read_files(Game *game, filetype_t ft) {
         rom.merge = stmt->get_string("merge");
         rom.status = static_cast<Rom::Status>(stmt->get_int("status"));
         rom.where = static_cast<where_t>(stmt->get_int("location"));
+        rom.mia = stmt->get_bool("missing");
         rom.hashes = stmt->get_hashes();
         rom.hashes.size = stmt->get_uint64("size", Hashes::SIZE_UNKNOWN);
 
@@ -449,7 +460,8 @@ void RomDB::read_files(Game *game, filetype_t ft) {
 std::vector<std::string> RomDB::read_list(enum dbh_list type) {
     static const std::unordered_map<enum dbh_list, Statement> query_list = {
         { DBH_KEY_LIST_DISK, QUERY_LIST_DISK },
-        { DBH_KEY_LIST_GAME, QUERY_LIST_GAME }
+        { DBH_KEY_LIST_GAME, QUERY_LIST_GAME },
+        { DBH_KEY_LIST_MIA, QUERY_LIST_MIA }
     };
     
     auto it = query_list.find(type);
@@ -636,7 +648,7 @@ void RomDB::write_game(Game *game) {
     }
     catch (Exception &e) {
         delete_game(game);
-        throw e;
+        throw;
     }
 }
 
@@ -654,6 +666,7 @@ void RomDB::write_files(Game *game, filetype_t ft) {
         stmt->set_string("merge", rom.merge);
         stmt->set_int("status", rom.status);
         stmt->set_int("location", rom.where);
+        stmt->set_bool("missing", rom.mia);
         stmt->set_uint64("size", rom.hashes.size, Hashes::SIZE_UNKNOWN);
         stmt->set_hashes(rom.hashes, true);
         
@@ -711,3 +724,11 @@ size_t RomDB::get_detector_id_for_dat(size_t dat_no) const {
     // TODO: Fix once we support multiple detectors in one DB.
     return detectors.begin()->first;
 }
+
+bool RomDB::game_exists(const std::string &name) {
+    auto stmt = get_statement(QUERY_GAME_ID);
+
+    stmt->set_string("name", name);
+
+    return stmt->step();
+}
